/*
 * Copyright (c) 2010, Franziska Hufsky
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the Friedrich-Schiller-University Jena nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package de.unijena.bioinf.labelingdetector.ui;

import appcore.application.AbstractApplication;
import appcore.application.ApplicationComponent;
import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.layout.FormLayout;
import de.unijena.bioinf.labelingdetector.*;
import de.unijena.bioinf.labelingdetector.ui.wizard.SelectImporterPanel;
import epos.ui.util.ComponentUtils;
import epos.ui.util.EposTask;
import epos.ui.util.FileDialogs;
import epos.ui.util.components.ElementTablePanel;
import epos.ui.view.Toolbar;
import epos.ui.wizard.EposWizardComponenetFactory;
import epos.ui.wizard.WizardDialog;
import org.jdesktop.application.Action;
import org.jdesktop.application.Application;
import org.jdesktop.application.ApplicationActionMap;
import org.jdesktop.application.Task;

import javax.swing.*;
import java.awt.*;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.*;
import java.util.List;

/**
 * Date: 28-Jun-2010
 *
 * @author Thasso Griebel (thasso.griebel@gmail.com)
 */
public class LabelingDetectorApplication extends AbstractApplication<List<Spectrum>>{
    private JPanel body;
    private MoleculeSettingsPanel moleculeSettings;
    private AlgorithmSettingsPanel algorithmSettings;
    private MSSettingsPanel msSettings;
    private ElementTablePanel<AnalysedSpectrum> spectraPanel;

    @Override
    public JComponent getBody() {
        if(body == null){
            // create main component

            spectraPanel = new ElementTablePanel<AnalysedSpectrum>(new ElementTablePanel.AddElementsAction() {
                public void addElements(ElementTablePanel elementListPanel) {
                    // plus button click
                    addSpectrum();
                }
            }, new AnalysedSpectrumTableFormat());
            spectraPanel.setDragText("Input Spectra");
            spectraPanel.removeDraggableWidget();
            moleculeSettings = new MoleculeSettingsPanel();
            algorithmSettings = new AlgorithmSettingsPanel();
            msSettings = new MSSettingsPanel();



            JPanel settingsPanel = new JPanel();
            FormLayout fl = new FormLayout("f:p:g,2dlu,f:p:g,2dlu,f:p:g","t:p:g");
            DefaultFormBuilder builder = new DefaultFormBuilder(fl, settingsPanel);
            builder.append(moleculeSettings.getComponent(),algorithmSettings.getComponent(),msSettings.getComponent());


            // ui tweaks
            spectraPanel.setBorder(ComponentUtils.createInsetBorder(1,0,0,0));

            body = new JPanel(new BorderLayout());
            body.add(settingsPanel, BorderLayout.NORTH);
            body.add(spectraPanel, BorderLayout.CENTER);
        }
        return body;
    }


    @org.jdesktop.application.Action
    public void addSpectrum(){
        WizardDialog wizard = new WizardDialog(new EposWizardComponenetFactory());
        SelectImporterPanel importerSelect = new SelectImporterPanel();
        wizard.add(importerSelect);
        int i = wizard.show(body, "Enter spectra");
        if(i == WizardDialog.CANCELED) return;


        List<AnalysedSpectrum> spectrums = importerSelect.getSpectra();
        if (spectrums==null){
            final File file = importerSelect.getFile();
            new EposTask<List<AnalysedSpectrum>>(body,"Read file"){
                @Override
                protected List<AnalysedSpectrum> doInBackground() throws Exception {
                    List<AnalysedSpectrum> spectra = Utils.readInputSpectra(file);
                    return spectra;  //To change body of implemented methods use File | Settings | File Templates.
                }

                @Override
                protected void succeeded(List<AnalysedSpectrum> result) {
                    super.succeeded(result);
                    int c = spectraPanel.getElements().size();
                    for (AnalysedSpectrum spectrum : result) {
                        //AnalysedSpectrum a = new AnalysedSpectrum("Spectrum " + c++,spectrum );
                        spectraPanel.addData(Arrays.asList(spectrum));
                    }


                }
            }.execute();
        }else {
            int c = spectraPanel.getElements().size();
            for (AnalysedSpectrum spectrum : spectrums) {
               // AnalysedSpectrum a = new AnalysedSpectrum("Spectrum " + c++,spectrum );
                spectraPanel.addData(Arrays.asList(spectrum));    
            }

        }
    }

    @Action
    public void saveResults(){
        final File file = FileDialogs.showSaveDialog(body);
        if (file!=null){
            new EposTask(body, "Save"){
                   @Override
                    protected Object doInBackground() throws Exception {
                        String output="";
                       //header
                       output+="formula: "+moleculeSettings.getSumformula()+"\n";
                       if(!moleculeSettings.getModification().equals("")){
                        output+="unlabeled part: "+moleculeSettings.getModification()+"\n";
                       }
                       output+="charge: "+moleculeSettings.getCharge()+"\nC13/C12 ratio: ";
                       switch (moleculeSettings.getCarbonRatio().getCarbonRatio()){
                           case bio:    output+="biological (1.096%)\n"; break;
                           case carbonate:  output+="carbonate specific (1.112%)\n"; break;
                       }
                       output+="labeling: "+moleculeSettings.getLabeling().getLabeling()+"\n\n";

                       output+="scoring: ";
                       switch (algorithmSettings.getScoring().getScore()){
                           case MI:     output+="masses and intensities\n";break;
                           case MDI:    output+="mass differences and intensities\n";break;
                           case I:      output+="intensities\n";break;
                       }
                       if (algorithmSettings.getNormalizationOffset()>0){
                            output+="renormalize with peaks higher than: "+algorithmSettings.getNormalizationOffset()*100+"%\n";
                       }
                       if (algorithmSettings.getScoringOffset()>0){
                           output+="scored only peaks higher than: "+algorithmSettings.getScoringOffset()*100+"%\n";
                       }
                       output+="\n";

                       output+="alpha1="+msSettings.getAlpha1()+" alpha0="+msSettings.getAlpha0()+" beta1="+msSettings.getBeta1()+" beta0="+msSettings.getBeta0()+"\n";
                       output+="\n---------------------------------------------------------------------------\n\n";
                       //results
                        for (AnalysedSpectrum a : spectraPanel.getElements() ) {
                            output+=a.getId()+"\t"+a.getInfo()+"\t"+a.getValue()+"\n";
                        }

                        try {
                            BufferedWriter out = new BufferedWriter(new FileWriter(file));
                            out.write(output);
			                out.close();
                        } catch (IOException e) {
                            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                        }

                        return null;  //To change body of implemented methods use File | Settings | File Templates.
                    }
            }.execute();

        }

    }

    @Override
    public ApplicationComponent getMenu(ApplicationComponent.Position position) {
        if(position == ApplicationComponent.Position.TOP){
            // create Top toolbar
            ApplicationActionMap actions = Application.getInstance().getContext().getActionMap(this);

            return new Toolbar.Builder()
                    .add(actions.get("runAlgorithm"))
                    .add(actions.get("saveResults"))
                    .getToolbar();
        }
        return super.getMenu(position);
    }

    @Action
    public Task runAlgorithm(){



        return new EposTask<Map<AnalysedSpectrum, List<Double>>>(body, "Compute labeling"){
            @Override
            protected Map<AnalysedSpectrum, List<Double>> doInBackground() throws Exception {
                List<AnalysedSpectrum> spectra = spectraPanel.getElements();
                Map<AnalysedSpectrum, List<Double>> results = new HashMap<AnalysedSpectrum, List<Double>>();
                List<Spectrum> spcs = new ArrayList<Spectrum>();
                for (AnalysedSpectrum analysedSpectrum : spectra) {
                   spcs.add(analysedSpectrum.getSpectrum());
                }

                if (moleculeSettings.getSumformula().equals("")||moleculeSettings.getSumformula().contains("Enter")){
                    moleculeSettings.setSumformula("Enter Sumformula");
                } else{
                    AlgorithmSettings algoSets = new AlgorithmSettings(algorithmSettings.getNumberOfDecimal(), algorithmSettings.getScoring().getScore(), algorithmSettings.getScoringOffset(), algorithmSettings.getNormalizationOffset());
                    MSSettings msSets = new MSSettings(msSettings.getAlpha0(), msSettings.getAlpha1(), msSettings.getBeta0(), msSettings.getBeta1());
                    MoleculeSettings moleculeSets = new MoleculeSettings(moleculeSettings.getCharge(), moleculeSettings.getLabeling().getLabeling(), moleculeSettings.getConvertedMolecule(), moleculeSettings.getCarbonRatio().getCarbonRatio());
                    Analyzer analyzer = new Analyzer(algoSets,moleculeSets, msSets);
                    for (int i = 0; i < spcs.size(); i++) {
                        setMessage("Analysing " + i);
                        double[] d = analyzer.runAnalysis(spcs.get(i));
                        List<Double> dlist = new ArrayList<Double>();
                        dlist.add(d[0]);
                        dlist.add(d[1]);
                        results.put(spectra.get(i), dlist);
                        if(Thread.interrupted()) return null;
                    }

                }
                return results;

            }

            @Override
            protected void succeeded(Map<AnalysedSpectrum, List<Double>> result) {
                super.succeeded(result);
                for (Map.Entry<AnalysedSpectrum, List<Double>> e : result.entrySet()) {
                    e.getKey().setValue(e.getValue().get(0));
                    e.getKey().setSumSquare(e.getValue().get(1));
                }
                spectraPanel.revalidate();
            }
        };
    }
}
